/*******************************************************************************

 WINBINDER - A native Windows binding

 Copyright  2004 Hypervisual - see LICENSE.TXT for details
 Author: Rubem Pechansky (pechansky@hypervisual.com)

 Functions for Windows controls

*******************************************************************************/

//----------------------------------------------------------------- DEPENDENCIES

#include "wb.h"
#include <stdlib.h>				// For itoa(), atoi()
#include <commctrl.h>

//-------------------------------------------------------------------- CONSTANTS

#define MAXWINCLASS		32

//-------------------------------------------------------------------- VARIABLES

// Global

WNDPROC lpfnTabProcOld = NULL;
WNDPROC lpfnFrameProcOld = NULL;

//---------------------------------------------------------- FUNCTION PROTOTYPES

// Private

static BOOL SetTransparentBitmap(HWND hwnd, HBITMAP hbmBits, BOOL bStatic, COLORREF clTransp);
static LRESULT CALLBACK FrameProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

//----------------------------------------------------------- EXPORTED FUNCTIONS

/*
	TODO: Call wbCreateToolbar() if wbclass is a toolbar
	TODO: Call wbCreateMenu() if wbclass is a menu
	TODO: Add container to tabs and/or remove background from the
		following controls (at least): Label, CheckBox, Slider. Background is
		visible when they are over a tab control on themed Windows XP.
	TODO: Subclass static controls to add background over tabs
*/

PWBOBJ wbCreateControl(PWBOBJ pwboParent, UINT uWinBinderClass, LPCTSTR pszCaption,
  int xPos, int yPos, int nWidth, int nHeight, UINT id, DWORD dwWBStyle, long lParam)
{
	PWBOBJ pwbo;
	LPTSTR pszClass;
	int nVisible;
	DWORD dwStyle, dwExStyle;
	HWND hTab;

	if(!pwboParent || !pwboParent->handle || !IsWindow(pwboParent->handle)) {
		wbError(MB_ICONWARNING, "Cannot create control if parent window is NULL");
		return NULL;
	}

	pwbo = wbAlloc(sizeof(WBOBJ));
	if(!pwbo)
		return NULL;

	pwbo->handle = NULL;
	pwbo->id = id;
	pwbo->uClass = uWinBinderClass;
	pwbo->item = -1;
	pwbo->subitem = -1;
	pwbo->style = dwWBStyle;
	pwbo->parent = pwboParent;
	pwbo->pszCallBackFn = NULL;
	pwbo->lparam = lParam;

	// If parent is a tab control, change the parent to a page handle

	if(pwboParent->uClass == TabControl) {

		PTABDATA pTabData;

		pTabData = (PTABDATA)pwboParent->lparam;

		// pwbo->lparam is the page index

		hTab = pTabData->page[pwbo->lparam].hwnd ?
		  pTabData->page[pwbo->lparam].hwnd : pwboParent->handle;

	}

	// Set visible / enabled flags

	nVisible = BITTEST(dwWBStyle, WBC_INVISIBLE) ? 0 : WS_VISIBLE;
	nVisible |= BITTEST(dwWBStyle, WBC_DISABLED) ? WS_DISABLED : 0;

	if(BITTEST(dwWBStyle, WBC_BORDER))
		dwExStyle = WS_EX_STATICEDGE;
	else
		dwExStyle = 0;

	// Set Windows class and styles according to WinBinder class

	switch(uWinBinderClass) {

		//-------------------------------- Regular child controls

		case Label:
			pszClass = "STATIC";
			if(BITTEST(dwWBStyle, WBC_ELLIPSIS))
				dwStyle = WS_CHILD | SS_WORDELLIPSIS | nVisible;
			else
				dwStyle = WS_CHILD | nVisible;

			if(BITTEST(dwWBStyle, WBC_CENTER))
				dwStyle |= SS_CENTER;
			else if(BITTEST(dwWBStyle, WBC_RIGHT))
				dwStyle |= SS_RIGHT;
			else {
				dwStyle |= SS_LEFT;
				if(!BITTEST(dwWBStyle, WBC_LINES))
					dwStyle |= SS_LEFTNOWORDWRAP;
			}
			break;

		case Frame:
			if((nWidth > nHeight) && nHeight <= 8) {			// Horizontal line
				pszClass = "STATIC";
				dwStyle = WS_CHILD | SS_ETCHEDHORZ | nVisible;
			} else if((nWidth < nHeight) && nWidth <= 8) {		// Vertical line
				pszClass = "STATIC";
				dwStyle = WS_CHILD | SS_ETCHEDVERT | nVisible;
			} else if(BITTEST(dwWBStyle, WBC_IMAGE)) {			// Image
				pszClass = "STATIC";
				dwStyle = WS_CHILD | SS_WORDELLIPSIS | SS_CENTERIMAGE | nVisible;
				if(BITTEST(dwWBStyle, WBC_CENTER))
					dwStyle |= SS_CENTER;
				else if(BITTEST(dwWBStyle, WBC_RIGHT))
					dwStyle |= SS_RIGHT;
				else
					dwStyle |= SS_LEFT;
			} else {											// Group box
				pszClass = "BUTTON";
				dwStyle = WS_CHILD | BS_GROUPBOX | nVisible;
				dwExStyle |= WS_EX_TRANSPARENT;
			}
			break;

		case PushButton:
			pszClass = "BUTTON";
			dwStyle = WS_CHILD | WS_TABSTOP | BS_PUSHBUTTON | BS_NOTIFY | nVisible;
			break;

		case CheckBox:
			pszClass = "BUTTON";
			dwStyle = WS_CHILD | WS_TABSTOP | BS_AUTOCHECKBOX | BS_NOTIFY | nVisible;
			break;

		case RadioButton:
			pszClass = "BUTTON";
			dwStyle = WS_CHILD | WS_TABSTOP | BS_AUTORADIOBUTTON | BS_NOTIFY | nVisible;
			if(BITTEST(dwWBStyle, WBC_GROUP))
				dwStyle |= WS_GROUP;
			break;

		case EditBox:
			pszClass = "EDIT";

			dwStyle = WS_CHILD | WS_TABSTOP | nVisible;
			if(BITTEST(dwWBStyle, WBC_READONLY))
				dwStyle |= ES_READONLY;
			if(BITTEST(dwWBStyle, WBC_MASKED))
				dwStyle |= ES_PASSWORD;
			if(BITTEST(dwWBStyle, WBC_NUMBER))
				dwStyle |= ES_NUMBER;
			if(BITTEST(dwWBStyle, WBC_LINES)) {
				if(BITTEST(dwWBStyle, WBC_READONLY))
					dwStyle |= WS_VSCROLL | ES_MULTILINE;
				else
					dwStyle |= WS_VSCROLL | ES_MULTILINE | ES_WANTRETURN;
			} else {
				dwStyle |= ES_AUTOHSCROLL;
				if(BITTEST(dwWBStyle, WBC_CENTER))
					dwStyle |= ES_CENTER;
				else if(BITTEST(dwWBStyle, WBC_RIGHT))
					dwStyle |= ES_RIGHT;
				else
					dwStyle |= ES_LEFT;
			}
			dwExStyle |= WS_EX_CLIENTEDGE;
			break;

		case RTFEditBox:
			pszClass = RICHEDITCONTROL;
			dwStyle = WS_CHILD | WS_TABSTOP | nVisible;
			if(BITTEST(dwWBStyle, WBC_READONLY))
				dwStyle |= ES_READONLY | WS_VSCROLL | ES_MULTILINE;
			else
				dwStyle |= WS_VSCROLL | ES_MULTILINE | ES_WANTRETURN;
			if(BITTEST(dwWBStyle, WBC_CENTER))
				dwStyle |= ES_CENTER;
			else if(BITTEST(dwWBStyle, WBC_RIGHT))
				dwStyle |= ES_RIGHT;
			else
				dwStyle |= ES_LEFT;
			dwExStyle |= WS_EX_CLIENTEDGE;
			break;

		case ListBox:
			pszClass = "LISTBOX";
			dwStyle = WS_CHILD | WS_TABSTOP | WS_VSCROLL | LBS_NOTIFY | nVisible;
			if(BITTEST(dwWBStyle, WBC_SORT))
				dwStyle |= LBS_SORT;
			dwExStyle |= WS_EX_CLIENTEDGE;
			break;

		case ComboBox:
			pszClass = "COMBOBOX";
			dwStyle = WS_CHILD | WS_TABSTOP | WS_VSCROLL | CBS_DROPDOWNLIST | nVisible;
			if(BITTEST(dwWBStyle, WBC_SORT))
				dwStyle |= CBS_SORT;
			break;

		case ScrollBar:
			pszClass = "SCROLLBAR";
			if(nHeight >= nWidth)
				dwStyle = WS_CHILD | WS_TABSTOP | SBS_VERT | nVisible;
			else
				dwStyle = WS_CHILD | WS_TABSTOP | nVisible;
			break;

		case ListView:
			pszClass = WC_LISTVIEW;
			dwStyle = WS_CHILD | WS_TABSTOP | LVS_REPORT | LVS_SHOWSELALWAYS | LVS_NOLABELWRAP | nVisible;
			if(!BITTEST(dwWBStyle, WBC_SORT))
				dwStyle |= LVS_NOSORTHEADER;
			dwExStyle |= WS_EX_CLIENTEDGE;
			break;

		case Spinner:
			pszClass = UPDOWN_CLASS;
			dwStyle = WS_CHILD | WS_TABSTOP | UDS_ARROWKEYS | UDS_AUTOBUDDY | UDS_SETBUDDYINT | UDS_NOTHOUSANDS	| nVisible;
			break;

		case Gauge:
			pszClass = PROGRESS_CLASS;
			dwStyle = WS_CHILD | nVisible;
			break;

		case Slider:
			pszClass = TRACKBAR_CLASS;
			if(BITTEST(dwWBStyle, WBC_LINES))
				dwStyle = WS_CHILD | WS_TABSTOP | TBS_AUTOTICKS | nVisible;
			else
				dwStyle = WS_CHILD | WS_TABSTOP | TBS_NOTICKS | nVisible;
			break;

		case TreeView:
			{
				PTREEDATA ptrdt;
				int i;

				pszClass = WC_TREEVIEW;
				dwStyle = WS_CHILD | WS_TABSTOP | TVS_HASBUTTONS | TVS_LINESATROOT | nVisible;
				dwExStyle |= WS_EX_CLIENTEDGE;

				// Attach a structure to the treeview control and initialize it

				ptrdt = wbAlloc(sizeof(TREEDATA));
//				memset(ptrdt, 0, sizeof(TREEDATA));

				ptrdt->hLast = TVI_ROOT;
				for(i = 0; i < MAX_TREEVIEW_LEVELS; i++) {
					ptrdt->hParent[i] = TVI_ROOT;
					ptrdt->bCustomImages[i] = FALSE;
				}
				ptrdt->nLastLevel = 0;
				pwbo->lparam = (LPARAM)ptrdt;
			}
			break;

		case StatusBar:
			pszClass =  STATUSCLASSNAME;
			if(BITTEST(GetWindowLong(pwboParent->handle, GWL_STYLE), WS_THICKFRAME))
				dwStyle = WS_CHILD | CCS_BOTTOM | CCS_NOMOVEY | SBARS_SIZEGRIP | nVisible;
			else
				dwStyle = WS_CHILD | CCS_BOTTOM | CCS_NOMOVEY | nVisible;
			break;

		case TabControl:
			pszClass = WC_TABCONTROL;
			dwStyle = WS_CHILD | WS_CLIPCHILDREN | TCS_FOCUSNEVER | nVisible;
//			dwStyle = WS_CHILD | WS_TABSTOP | WS_CLIPCHILDREN | nVisible;
			{
				PTABDATA pTabData;

				// Alloc memory for tab data

				pTabData = wbAlloc(sizeof(TABDATA));
				memset(pTabData, 0, sizeof(TABDATA));
				pwbo->lparam = (LONG)pTabData;
				pwbo->uClass = TabControl;

			}
			break;

		default:
			wbError(MB_ICONWARNING, "Unknown control class %d", uWinBinderClass);
			return NULL;

	} // end switch

	// Create the control

	pwbo->handle = CreateWindowEx(
		dwExStyle,
		pszClass,
		pszCaption && *pszCaption ? pszCaption : "",
		dwStyle,
		xPos, yPos, nWidth, nHeight,
		(pwboParent->uClass == TabControl) ? hTab : (pwboParent ? pwboParent->handle : NULL),
		(HMENU)id,
		hAppInstance,
		NULL
	);

	if(!pwbo->handle) {
		wbError(MB_ICONWARNING, "Error creating control");
		return NULL;
	}

	// Set control font

	SendMessage(pwbo->handle, WM_SETFONT, (WPARAM)hIconFont, 0);

	// Further processing after control creation

	switch(pwbo->uClass) {

		case TabControl:		// Subclasses tab control to process WM_COMMAND
			lpfnTabProcOld = (WNDPROC)SetWindowLong(pwbo->handle, GWL_WNDPROC, (LONG)TabProc);
			wbSetTabControlText(pwbo, pszCaption);

			// Assign window handler, if any

			if(pwbo->parent && pwbo->parent->pszCallBackFn && *pwbo->parent->pszCallBackFn)
				pwbo->pszCallBackFn = pwbo->parent->pszCallBackFn;
			break;

		case Frame:				// Subclasses Frame control to process WM_COMMAND
			if(!stricmp(pszClass, "BUTTON"))		// Only group boxes!
				lpfnFrameProcOld = (WNDPROC)SetWindowLong(pwbo->handle, GWL_WNDPROC, (LONG)FrameProc);
			break;

		case Spinner:
			SendMessage(pwbo->handle, UDM_SETRANGE, 0, MAKELONG(100, 0));
			SendMessage(pwbo->handle, UDM_SETPOS, 0, MAKELONG((short)0, 0));
			break;

		case ScrollBar:
			SendMessage(pwbo->handle, SBM_SETRANGE, 0, 100);
			break;

		case StatusBar:
			SetStatusBarId(id);
			break;

		case Slider:
			if(BITTEST(dwWBStyle, WBC_LINES))
				SendMessage(pwbo->handle, TBM_SETTICFREQ, 10, 1);
			break;

		case ListView:
			SendMessage(pwbo->handle, LVM_SETEXTENDEDLISTVIEWSTYLE, 0, LVS_EX_FULLROWSELECT | LVS_EX_HEADERDRAGDROP | LVS_EX_LABELTIP | LVS_EX_SUBITEMIMAGES);
			if(BITTEST(dwWBStyle, WBC_LINES))
				SendMessage(pwbo->handle, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_GRIDLINES, LVS_EX_GRIDLINES);
			if(BITTEST(dwWBStyle, WBC_CHECKBOXES))
				SendMessage(pwbo->handle, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_CHECKBOXES, LVS_EX_CHECKBOXES);
			break;

		case RTFEditBox:
			SendMessage(pwbo->handle, EM_FMTLINES, TRUE, 0);
			SendMessage(pwbo->handle, EM_SETEVENTMASK, 0, ENM_CHANGE);	// Enables EN_CHANGE messages. Should be EN_UPDATE instead?
			break;
	}

	// Is the parent window a tab control?

	if(pwboParent->uClass == TabControl)
		RegisterControlInTab(pwboParent, pwbo, id, lParam);

	SetWindowLong(pwbo->handle, GWL_USERDATA, (LONG)pwbo);

	return pwbo;
}

/*
	TODO: Process treeview items?
	TODO: Must validate toolbar item first
*/

PWBOBJ wbGetControl(PWBOBJ pwboParent, int id)
{
	PWBOBJ pwbo;

	if(!wbIsWBObj(pwboParent)) { 						// Is it a valid control?
		wbError(MB_ICONWARNING, "%s: First parameter is not a valid WinBinder control", __func__);
		return NULL;
	}

	if(pwboParent->uClass == Menu) {

		if(IsMenu((HMENU)pwboParent->handle)) {		// Is it a menu?

			MENUITEMINFO mi;

			mi.cbSize = sizeof(MENUITEMINFO);
			mi.fMask = MIIM_ID;
			if(GetMenuItemInfo((HMENU)pwboParent->handle, id, FALSE, &mi)) {

				pwbo = wbAlloc(sizeof(WBOBJ));
				memcpy(pwbo, pwboParent, sizeof(WBOBJ));
				pwbo->id = id;

			} else {

				wbError(MB_ICONWARNING, "Menu item #%d not found in menu #%d", id, (int)pwboParent->handle);
				return NULL;
			}
		} else {
			wbError(MB_ICONWARNING, "Menu object does not have a valid menu handle");
			return NULL;
		}

	} else if(!IsWindow(pwboParent->handle)) {			// Is it a window?

		wbError(MB_ICONWARNING, "Object does not have a valid window handle");
		return NULL;

	} else {											// Yes, it's a window

		switch(pwboParent->uClass) {

			case ToolBar:								// It's a toolbar

				pwbo = wbAlloc(sizeof(WBOBJ));
				pwbo->handle = pwboParent->handle;
				pwbo->id = id;
				pwbo->uClass = pwboParent->uClass;
				pwbo->item = -1;
				pwbo->subitem = -1;
				pwbo->style = 0;
				pwbo->parent = pwboParent;
				pwbo->pszCallBackFn = NULL;
				pwbo->lparam = 0;
				break;

			default:									// Regular window classes
				{
					HWND hCtrl;

					hCtrl = GetDlgItem(pwboParent->handle, id);

					if(hCtrl) {

						// It's a regular control

						if(IsWindow(hCtrl)) {
							pwbo = (PWBOBJ)GetWindowLong(hCtrl, GWL_USERDATA);
						} else {
							pwbo = wbAlloc(sizeof(WBOBJ));
							pwbo->handle = hCtrl;
							pwbo->id = id;
							pwbo->uClass = pwboParent->uClass;
							pwbo->item = -1;
							pwbo->subitem = -1;
							pwbo->style = 0;
							pwbo->parent = pwboParent;
							pwbo->pszCallBackFn = NULL;
							pwbo->lparam = 0;
						}

					} else {

						// Is its parent a tab control?

						if(pwboParent->uClass == TabControl) {

							// Yes, let's try to find it

							PTABDATA pTabData;
							UINT i;

							pTabData = (PTABDATA)pwboParent->lparam;

							for(i = 0; i < pTabData->nCtrls; i++) {
								if(pTabData->ctrl[i].id == (UINT)id) {

									hCtrl = pTabData->ctrl[i].hwnd;

									if(IsWindow(hCtrl)) {
										pwbo = (PWBOBJ)GetWindowLong(hCtrl, GWL_USERDATA);
									} else {
										pwbo = NULL;
									}
									break;
								}
							}

						} else {

							// No, control does not exist

							char szText[256];

							wbGetText(pwboParent, szText, 256);
							wbError(MB_ICONWARNING, "Control #%d not found in window '%s'", id, szText);
							return NULL;
						}
					}
				}
		} // switch
	} // else

	return pwbo;
}

BOOL wbSetText(PWBOBJ pwbo, LPCTSTR pszText)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	switch(pwbo->uClass) {

		case Spinner:

			// Change text of "buddy" control. The buddy is the edit box attached to the Up-down control
			{
				HWND hctrl = (HWND)SendMessage(pwbo->handle, UDM_GETBUDDY, 0, 0);
				return SetWindowText(hctrl, pszText);
			}
			break;

		case TabControl:

			return wbSetTabControlText(pwbo, pszText);

		default:

			return SetWindowText(pwbo->handle, pszText);
	}
}

/*
	Limitation: Treeview items <= 1024 characters
*/

UINT wbGetTextLength(PWBOBJ pwbo)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	switch(pwbo->uClass) {

		case Spinner:
			{
				// Measure length of text of "buddy" control. The buddy is the edit box attached to the Up-down control
				HWND hBuddy;

				hBuddy = (HWND)SendMessage(pwbo->handle, UDM_GETBUDDY, 0, 0);
				return GetWindowTextLength(hBuddy);
			}

		case ListBox:
			{
				int nIndex;

				nIndex = SendMessage(pwbo->handle, LB_GETCURSEL, 0, 0);
				if(nIndex == LB_ERR)
					return 0;
				return SendMessage(pwbo->handle, LB_GETTEXTLEN, nIndex, 0);
			}

		case ComboBox:
			{
				int nIndex;

				nIndex = SendMessage(pwbo->handle, CB_GETCURSEL, 0, 0);
				if(nIndex == CB_ERR)
					return 0;
				return SendMessage(pwbo->handle, CB_GETLBTEXTLEN, nIndex, 0);
			}

		case TreeView:
			{
				TV_ITEM tvi;
				char szText[MAX_ITEM_STRING];

				tvi.hItem = TreeView_GetSelection(pwbo->handle);
				if(!tvi.hItem)
					return 0;

				tvi.pszText = szText;
				tvi.cchTextMax = MAX_ITEM_STRING;
				tvi.mask = TVIF_TEXT;
				TreeView_GetItem(pwbo->handle, &tvi);
				if(tvi.pszText && *tvi.pszText)
					return strlen(tvi.pszText);
				else
					return 0;
			}

		default:
			/*
				TODO: Trying to fix control text length bug

				According to WinHelp for GetWindowTextLength, "To obtain the exact length of
				the text, use (...) the GetWindowText function." Makes no difference, however.
			{
				char *pszText;
				int nLen;

				pszText = wbAlloc(32767);
				nLen = GetWindowText(pwbo->handle, pszText, 32767);
				wbFree(pszText);
				return nLen;
			}*/
			return GetWindowTextLength(pwbo->handle);
	}
}

BOOL wbGetText(PWBOBJ pwbo, LPTSTR pszText, UINT nMaxChars)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	if(!nMaxChars)
		return FALSE;

	*pszText = '\0';

	switch(pwbo->uClass) {

		case Spinner:
			{
				// Reads text of "buddy" control. The buddy is the edit box attached to the Up-down control
				HWND hBuddy;

				hBuddy = (HWND)SendMessage(pwbo->handle, UDM_GETBUDDY, 0, 0);
				GetWindowText(hBuddy, pszText, nMaxChars);
				return TRUE;
			}

		case ListBox:
			{
				int nIndex;

				nIndex = SendMessage(pwbo->handle, LB_GETCURSEL, 0, 0);
				if(nIndex == LB_ERR)
					return FALSE;
				return (SendMessage(pwbo->handle, LB_GETTEXT, nIndex, (LPARAM)(LPTSTR)pszText) != LB_ERR);
			}

		case ComboBox:
			// TODO: The combos return a newline after the text
			{
				int nIndex;

				nIndex = SendMessage(pwbo->handle, CB_GETCURSEL, 0, 0);
				if(nIndex == CB_ERR)
					return FALSE;
				return (SendMessage(pwbo->handle, CB_GETLBTEXT, nIndex, (LPARAM)(LPTSTR)pszText) != CB_ERR);
			}

		case TreeView:

			{
				TV_ITEM tvi;

				tvi.hItem = TreeView_GetSelection(pwbo->handle);
				if(!tvi.hItem)
					return FALSE;
				tvi.pszText = pszText;
				tvi.cchTextMax = nMaxChars;
				tvi.mask = TVIF_TEXT;
				TreeView_GetItem(pwbo->handle, &tvi);
				return (tvi.pszText && *tvi.pszText);
			}

		default:

			GetWindowText(pwbo->handle, pszText, nMaxChars);
			return TRUE;
	}
}

BOOL wbSetEnabled(PWBOBJ pwbo, BOOL bState)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	if(IsMenu((HMENU)pwbo->handle)) {		// Is it a menu item?

		MENUITEMINFO mi;
		BOOL bRet;

		mi.cbSize = sizeof(MENUITEMINFO);
		mi.fMask = MIIM_STATE;
		mi.fState = bState ? MFS_ENABLED : (MFS_DISABLED | MFS_GRAYED);

		bRet = SetMenuItemInfo((HMENU)pwbo->handle, pwbo->id, FALSE, &mi);
		if(!bRet)
			wbError(MB_ICONWARNING, "Could not set state of menu item #%d", pwbo->id);
		return bRet;

	} else if(IsWindow(pwbo->handle)) {		// Is it a window?

		switch(pwbo->uClass) {

			case ToolBar:

				return SendMessage((HWND)pwbo->handle, TB_ENABLEBUTTON, pwbo->id, MAKELONG(bState, 0));
/*
			// Probably useless code: introduced in build 0.28.100
			// Must not return TRUE, must call EnableWindow() instead and return its value

			case EditBox:
			case RTFEditBox:
				{
					DWORD dwStyle;

					dwStyle = GetWindowLong(pwbo->handle, GWL_STYLE);
					if(!bState && (BITTEST(dwStyle, ES_MULTILINE | ES_WANTRETURN) ||
					  BITTEST(dwStyle, ES_MULTILINE | ES_READONLY)))
						SetWindowLong(pwbo->handle, GWL_STYLE, dwStyle & ~ES_WANTRETURN);
					else if(bState && BITTEST(dwStyle, ES_MULTILINE))
						SetWindowLong(pwbo->handle, GWL_STYLE, dwStyle | ES_WANTRETURN);

					return TRUE;
				}*/

			default:
				return EnableWindow(pwbo->handle, bState ? 1 : 0);
		}
	} else
		return FALSE;
}

BOOL wbGetEnabled(PWBOBJ pwbo)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	if(IsMenu((HMENU)pwbo->handle)) {		// Is it a menu item?

		MENUITEMINFO mi;

		mi.cbSize = sizeof(MENUITEMINFO);
		mi.fMask = MIIM_STATE;

		if(!GetMenuItemInfo((HMENU)pwbo->handle, pwbo->id, FALSE, &mi)) {
			wbError(MB_ICONWARNING, "Could not get state of menu item #%d.", pwbo->id);
			return FALSE;
		}
		return (mi.fState & (MFS_DISABLED | MFS_GRAYED) ? FALSE : TRUE);

	} else if(IsWindow(pwbo->handle)) {		// Is it a window?

		switch(pwbo->uClass) {
			case ToolBar:								// Toolbar
				return SendMessage(pwbo->handle, TB_ISBUTTONENABLED, pwbo->id, 0);
			default:
				return IsWindowEnabled(pwbo->handle);	// The window
		}
	} else
		return FALSE;
}

DWORD wbGetSelected(PWBOBJ pwbo)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	switch(pwbo->uClass) {

		case TabControl:
			return SendMessage(pwbo->handle, TCM_GETCURSEL, 0, 0);

		default:
			return wbGetValue(pwbo);
	}
}

DWORD wbGetItemCount(PWBOBJ pwbo)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	switch(pwbo->uClass) {

		case ListView:
			return SendMessage(pwbo->handle, LVM_GETITEMCOUNT, 0, 0);

		case ListBox:
			return SendMessage(pwbo->handle, LB_GETCOUNT, 0, 0);

		case ComboBox:
			return SendMessage(pwbo->handle, CB_GETCOUNT, 0, 0);

		default:
			return 0;
	}
}

BOOL wbClearItems(PWBOBJ pwbo)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	switch(pwbo->uClass) {

		case ListView:
			return SendMessage(pwbo->handle, LVM_DELETEALLITEMS, 0, 0);

		case ListBox:
			SendMessage(pwbo->handle, LB_RESETCONTENT, 0, 0);
			break;

		case ComboBox:
			SendMessage(pwbo->handle, CB_RESETCONTENT, 0, 0);
			break;

		case TreeView:
			{
				PTREEDATA ptrdt	= (PTREEDATA)pwbo->lparam;
				int i;

				// Clear the treeview data

				ptrdt = wbAlloc(sizeof(TREEDATA));

				ptrdt->hLast = TVI_ROOT;
				for(i = 0; i < MAX_TREEVIEW_LEVELS; i++) {
					ptrdt->hParent[i] = TVI_ROOT;
					ptrdt->bCustomImages[i] = FALSE;
				}
				ptrdt->nLastLevel = 0;
				pwbo->lparam = (LPARAM)ptrdt;
			}
			return TreeView_DeleteAllItems(pwbo->handle);
//			return SendMessage(pwbo->handle, TVM_DELETEITEM, 0, (LPARAM)TVI_ROOT);
	}
	return TRUE;
}

BOOL wbGetVisible(PWBOBJ pwbo)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	return IsWindowVisible(pwbo->handle);
}

BOOL wbSetVisible(PWBOBJ pwbo, BOOL bState)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	ShowWindow(pwbo->handle, bState ? SW_SHOW : SW_HIDE);
	return TRUE;
}

BOOL wbSetFocus(PWBOBJ pwbo)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	switch(pwbo->uClass) {

		case EditBox:
			SendMessage(pwbo->handle, EM_SETSEL, 0, -1);	// Select all
			break;
	}

	return (SetFocus(pwbo->handle) != NULL);
}

// This function is temporarily disabled

BOOL wbSetStyle(PWBOBJ pwbo, DWORD dwWBStyle)
{
/*	DWORD dwOldStyle, dwOldExStyle, dwStyle, dwExStyle;

	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	pwbo->style = dwWBStyle;
	dwOldStyle = GetWindowLong(pwbo->handle, GWL_STYLE);
	dwOldExStyle = GetWindowLong(pwbo->handle, GWL_EXSTYLE);

	if(BITTEST(dwWBStyle, WBC_CHECKBOXES)) {
		dwOldStyle = SendMessage(pwbo->handle, LVM_GETEXTENDEDLISTVIEWSTYLE, 0, 0);
		SendMessage(pwbo->handle, LVM_SETEXTENDEDLISTVIEWSTYLE, 0, dwOldStyle | LVS_EX_CHECKBOXES);
		return TRUE;
	}

		case Frame:
		case Label:
			if(BITTEST(dwWBStyle, WBC_ELLIPSIS))
				dwStyle = dwOldStyle | SS_WORDELLIPSIS;
			if(BITTEST(dwWBStyle, WBC_CENTER))
				dwStyle = dwOldStyle | SS_CENTER;
			else if(BITTEST(dwWBStyle, WBC_RIGHT))
				dwStyle = dwOldStyle | SS_RIGHT;
			else if(!BITTEST(dwWBStyle, WBC_LINES))
				dwExStyle = dwOldExStyle | SS_LEFTNOWORDWRAP;
			break;

		default:
			return FALSE;
	}*/
	return TRUE;
}

BOOL wbSetValue(PWBOBJ pwbo, DWORD dwValue)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	switch(pwbo->uClass) {

		// Set value

		case Slider:
			return SendMessage(pwbo->handle, TBM_SETPOS, TRUE, dwValue);

		case Gauge:
			return SendMessage(pwbo->handle, PBM_SETPOS, (WPARAM)dwValue, 0);

		case Spinner:
			return SendMessage(pwbo->handle, UDM_SETPOS, 0, MAKELONG((short)dwValue, 0));

		case ScrollBar:
			return SendMessage(pwbo->handle, SBM_SETPOS, (WPARAM)dwValue, TRUE);

		case ComboBox:
			return SendMessage(pwbo->handle, CB_SETCURSEL, (WPARAM)dwValue, 0);

		case ListBox:
			return SendMessage(pwbo->handle, LB_SETCURSEL, (WPARAM)dwValue, 0);

		// Set value as text

		case EditBox:
		case RTFEditBox:
			{
				char szText[32];

				itoa(dwValue, szText, 10);
				return SetWindowText(pwbo->handle, szText);
			}

		// Check / uncheck button

		case PushButton:
		case CheckBox:
		case RadioButton:
			return SendMessage(pwbo->handle, BM_SETCHECK, dwValue, 0);

		// Check/uncheck menu item

		case Menu:
			return wbSetMenuItemChecked(pwbo, dwValue);
	}

	return TRUE;
}

BOOL wbSetRange(PWBOBJ pwbo, DWORD dwMin, DWORD dwMax)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	switch(pwbo->uClass) {

		// Set range

		case Slider:
			return SendMessage(pwbo->handle, TBM_SETRANGE, 0, MAKELONG(dwMin, dwMax));

		case Gauge:
			return SendMessage(pwbo->handle, PBM_SETRANGE, 0, MAKELONG(dwMin, dwMax));

		case Spinner:		// Inverted so that the control increases upwards
			return SendMessage(pwbo->handle, UDM_SETRANGE, 0,
				(LPARAM)MAKELONG((short)dwMax, (short)dwMin));

		case ScrollBar:
			return SendMessage(pwbo->handle, SBM_SETRANGE, (short)dwMin, (short)dwMax);
	}

	return TRUE;
}

/*
	Value in buttons and check boxes: 0 means unchecked, 1 means checked.
*/

DWORD wbGetValue(PWBOBJ pwbo)
{
	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return 0;

	switch(pwbo->uClass) {

		case Slider:

			return SendMessage(pwbo->handle, TBM_GETPOS, 0, 0);

		case Spinner:

			return LOWORD(SendMessage(pwbo->handle, UDM_GETPOS, 0, 0));

		case EditBox:
		case RTFEditBox:
			{
				char szText[32];

				GetWindowText(pwbo->handle, szText, 31);
				return atoi(szText);
			}

		case ComboBox:

			return SendMessage(pwbo->handle, CB_GETCURSEL, 0, 0);

		case ListBox:

			return SendMessage(pwbo->handle, LB_GETCURSEL, 0, 0);

		case ScrollBar:

			return SendMessage(pwbo->handle, SBM_GETPOS, 0, 0);

		case PushButton:
		case CheckBox:
		case RadioButton:

			return (SendMessage(pwbo->handle, BM_GETCHECK, 0, 0) == BST_CHECKED);

		case TreeView:

			{
	    		TV_ITEM tvi;

				tvi.hItem = TreeView_GetSelection(pwbo->handle);
				if(!tvi.hItem)
					return 0;
				else {
					tvi.mask = TVIF_PARAM;
					TreeView_GetItem(pwbo->handle, &tvi);
					return tvi.lParam;
				}
			}

		case ListView:

			return wbGetListViewItemChecked(pwbo, pwbo->item);

		case Menu:
			return wbGetMenuItemChecked(pwbo);

		default:
//			wbError(MB_ICONWARNING, "Could not retrieve value of control #%d", pwbo->handle);
			return 0;
	}
}

/* NOTE: nIndex is not used */

BOOL wbSetImage(PWBOBJ pwbo, HANDLE hImage, UINT nIndex, COLORREF clTransp, LPARAM lParam)
{
	BOOL bRet = TRUE;
	LONG lStyle;

	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	if(!hImage)
		return FALSE;

	switch(pwbo->uClass) {

		case Menu:
			wbSetMenuItemImage(pwbo, hImage);
			break;

		case PushButton:
			lStyle = GetWindowLong(pwbo->handle, GWL_STYLE) & ~(BS_ICON | BS_BITMAP);
			bRet = TRUE;
			if(IsIcon(hImage)) {
				SetWindowLong(pwbo->handle, GWL_STYLE, lStyle | BS_ICON);
				SendMessage(pwbo->handle, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hImage);
				break;
			} else if(IsBitmap(hImage)) {
				SetWindowLong(pwbo->handle, GWL_STYLE, lStyle | BS_BITMAP);
				bRet = SetTransparentBitmap(pwbo->handle, (HBITMAP)hImage, FALSE, clTransp);
				break;
			} else
				bRet = FALSE;
			break;

		case Frame:

			lStyle = GetWindowLong(pwbo->handle, GWL_STYLE) & ~(SS_ICON | SS_BITMAP);
			bRet = TRUE;
			if(IsIcon(hImage)) {
				SetWindowLong(pwbo->handle, GWL_STYLE, lStyle | SS_ICON);
				SendMessage(pwbo->handle, STM_SETIMAGE, IMAGE_ICON, (LPARAM)hImage);
				InvalidateRect(pwbo->handle, NULL, FALSE);
				break;
			} else if(IsBitmap(hImage)) {
				SetWindowLong(pwbo->handle, GWL_STYLE, lStyle | SS_BITMAP);
				SetTransparentBitmap(pwbo->handle, (HBITMAP)hImage, TRUE, clTransp);
				InvalidateRect(pwbo->handle, NULL, FALSE);
				break;
			} else
				bRet = FALSE;
			break;

		case ListView:

			lParam = MAX(1, lParam);	// Image count
			bRet = wbCreateListViewImages(pwbo, hImage, lParam, clTransp);
			break;

		case TreeView:

			lParam = MAX(1, lParam);	// Image count
			bRet = wbCreateTreeViewImages(pwbo, hImage, lParam, clTransp);
			break;

		default:

			SendMessage(pwbo->handle, WM_SETICON, ICON_BIG, (WPARAM)hImage);
			bRet = TRUE;

	} // end switch
	return bRet;
}

BOOL wbCreateItem(PWBOBJ pwbo, LPCTSTR pszItemText)
{
	LRESULT lRet;

	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	if(!pszItemText || !*pszItemText)
		return FALSE;

	switch(pwbo->uClass) {

		case ListBox:
			lRet = SendMessage(pwbo->handle, LB_ADDSTRING, 0, (LPARAM)pszItemText);
			return ((lRet != LB_ERR) && (lRet != LB_ERRSPACE));

		case ComboBox:

			lRet = SendMessage(pwbo->handle, CB_ADDSTRING, 0, (LPARAM)pszItemText);
			if(lRet == CB_ERR || lRet == CB_ERRSPACE)
				return FALSE;
			lRet = SendMessage(pwbo->handle, CB_SETCURSEL, 0, 0);
			return (lRet != CB_ERR);

		case TabControl:
			return wbCreateTabItem(pwbo, pszItemText);

		default:
			return FALSE;
	}
}

/* Refreshes the control from window pwnd whose command is wCmd. */

BOOL wbRefreshControl(PWBOBJ pwbo, BOOL bNow)
{
	BOOL bRet;

	if(!wbIsWBObj(pwbo))					// Is it a valid control?
		return FALSE;

	if(!IsWindow(pwbo->handle))
		return FALSE;

	bRet = InvalidateRect(pwbo->handle, NULL, TRUE);
	if(bNow)
		bRet = UpdateWindow(pwbo->handle);
	return bRet;
}

//------------------------------------------- FUNCTIONS PUBLIC TO WINBINDER ONLY

BOOL IsIcon(HANDLE handle)
{
	ICONINFO ii;

	return GetIconInfo((HICON)handle, &ii);
}

BOOL IsBitmap(HANDLE handle)
{
	SIZE siz;

	return GetBitmapDimensionEx(handle, &siz);
}

//------------------------------------------------------------ PRIVATE FUNCTIONS

/*

For opaque bitmaps, set clTransp to CLR_INVALID

*/

static BOOL SetTransparentBitmap(HWND hwnd, HBITMAP hbmBits, BOOL bStatic, COLORREF clTransp)
{
	static HBITMAP hbm = NULL, hbmMask;
	HDC hdc;
	RECT rc;
	HBRUSH hbr;
	BITMAP bm;

//	if(clTransp == CLR_INVALID) {
//		SendMessage(hwnd, bStatic ? STM_SETIMAGE : BM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hbmBits);
//		return TRUE;
//	}

	// Create an empty bitmap the same size as hwnd

	GetClientRect(hwnd, &rc);

	//if(hbm)
		// DeleteObject(hbm);		// Resource leak, but must not delete hbm here
	hbm = wbCreateBitmap(rc.right - rc.left, rc.bottom - rc.top);
	hdc = wbCreateBitmapDC(hbm);

	// Paint hbmBits with a mask

	if(clTransp != CLR_INVALID)
		hbmMask = wbCreateMask(hbmBits, clTransp);
	hbr = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
	SelectObject(hdc, hbr);
	SelectObject(hdc, GetStockObject(WHITE_PEN));
	Rectangle(hdc, 0, 0, rc.right - rc.left, rc.bottom - rc.top);
	GetObject(hbmBits, sizeof(BITMAP), (LPVOID)&bm);
	if(clTransp == CLR_INVALID) {
		wbCopyPartialBits(hdc, hbmBits,
			((rc.right - rc.left) - bm.bmWidth) / 2,
			((rc.bottom - rc.top) - bm.bmHeight) / 2,
			rc.right - rc.left, rc.bottom - rc.top, 0, 0);
	} else {
		wbMaskPartialBits(hdc, hbmBits, hbmMask,
			((rc.right - rc.left) - bm.bmWidth) / 2,
			((rc.bottom - rc.top) - bm.bmHeight) / 2,
			rc.right - rc.left, rc.bottom - rc.top, 0, 0);
	}
	SendMessage(hwnd, bStatic ? STM_SETIMAGE : BM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hbm);

	// Cleanup

	DeleteObject(hbr);
	if(hbmMask && clTransp != CLR_INVALID)
		DeleteObject(hbmMask);
//	DeleteObject(hbm);				// Resource leak, but must not delete hbm here
	DeleteDC(hdc);

	return TRUE;
}

// Processing routine for Frame controls

static LRESULT CALLBACK FrameProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {

		case WM_COMMAND:		// Passes commands to parent window
		case WM_NOTIFY:
			{
				HWND hwndParent = GetParent(hwnd);

				if(hwndParent)
					SendMessage(hwndParent, msg, wParam, lParam);
			}
			break;
	}
	return CallWindowProc(lpfnFrameProcOld, hwnd, msg, wParam, lParam);
}

//------------------------------------------------------------------ END OF FILE
