/*******************************************************************************

 WINBINDER - A native Windows binding

 Copyright  2004 Hypervisual - see LICENSE.TXT for details
 Author: Rubem Pechansky (pechansky@hypervisual.com)

 Window creation and manipulation functions

*******************************************************************************/

//----------------------------------------------------------------- DEPENDENCIES

#include "wb.h"
#include <string.h>				// For stricmp()
#include <shellapi.h>			// For Shell_NotifyIcon()

//-------------------------------------------------------------------- CONSTANTS

#define MAXWINNAME				256
#define DEFAULT_WIN_STYLE		(WS_POPUP|WS_MINIMIZEBOX|WS_VISIBLE|WS_CLIPSIBLINGS|WS_CLIPCHILDREN|WS_MAXIMIZEBOX|WS_CAPTION|WS_BORDER|WS_SYSMENU|WS_THICKFRAME)

#define MODELESS_WINDOW_CLASS	"wbModelessDlg"
#define MODAL_WINDOW_CLASS		"wbModalDlg"
#define MAIN_WINDOW_CLASS		"wbMainWnd"
#define LAYOUT_WINDOW_CLASS		"wbLayoutDlg"

//----------------------------------------------------------------------- MACROS

#define CALL_CALLBACK(id,lp)	/* Call user function */ \
	{if(pwbobj && pwbobj->parent->pszCallBackFn && *pwbobj->parent->pszCallBackFn) { \
		wbCallUserFunction(pwbobj->parent->pszCallBackFn, pwbobj->parent, pwbobj, id, lp); }}

#define SEND_MESSAGE(m)			/* Send additional notification message */ \
	((pwbobj->parent->uClass == TabControl) ? (pwbobj->parent->parent->style & WBC_NOTIFY) && (pwbobj->parent->parent->lparam & (m)) : (pwbobj->parent->style & WBC_NOTIFY) && (pwbobj->parent->lparam & (m)))

//----------------------------------------------------------- PRIVATE PROTOTYPES

static HICON GetWindowIcon(HWND hwnd);
static BOOL AddTaskBarIcon(HWND hwnd);
static BOOL DeleteTaskBarIcon(HWND hwnd);
static void UpdateLVlParams(HWND hwnd);
static int CALLBACK CompareLVItemsAscending(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);
static int CALLBACK CompareLVItemsDescending(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);

// Window procedures

static LRESULT CALLBACK DefaultWBProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK LayoutWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK ModelessWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK ModalWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK TabPageProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

//-------------------------------------------------------------------- VARIABLES

// Global

HINSTANCE hAppInstance = NULL;
BOOL bMinimizeToTaskBar = FALSE;
HBRUSH hbrTabs = NULL;
COLORREF clrTabs = CLR_INVALID;

// Static

static int idStatusBar = 0;
static HWND hToolBar = NULL;
static HWND hwndListView = NULL;

//------------------------------------------------------------- PUBLIC FUNCTIONS

PWBOBJ wbCreateWindow(PWBOBJ pwboParent, UINT uWinBinderClass, LPCTSTR pszCaption, UINT xPos, UINT yPos, int nWidth, int nHeight, DWORD dwWinBinderStyle, long lParam)
{
	char szWindowName[MAXWINNAME] = "";
	LPTSTR pszClass;
	DWORD dwStyle = 0;
	DWORD dwExStyle = 0;
	PWBOBJ pwbo;

	// Initialize WinBinder object

	pwbo = wbAlloc(sizeof(WBOBJ));
	if(!pwbo)
		return NULL;

	pwbo->handle = NULL;
	pwbo->id = 0;
	pwbo->uClass = uWinBinderClass;
	pwbo->item = -1;
	pwbo->subitem = -1;
	pwbo->style = dwWinBinderStyle;
	pwbo->parent = pwboParent;
	pwbo->pszCallBackFn = NULL;
	pwbo->lparam = lParam;

	switch(uWinBinderClass) {

		case AppWindow:				// Fixed size main window
			pszClass = MAIN_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_CLIPCHILDREN | WS_MINIMIZEBOX;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_APPWINDOW;
			bMinimizeToTaskBar = (dwWinBinderStyle & WBC_COLLAPSED) != 0;
			break;

		case ResizableWindow:		// Resizable main window
			pszClass = MAIN_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPEDWINDOW | WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_CLIPCHILDREN | WS_SIZEBOX;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_APPWINDOW;
			bMinimizeToTaskBar = (dwWinBinderStyle & WBC_COLLAPSED) != 0;
			break;

		case PopupWindow:			// Fixed size main window, no minimize button
			pszClass = MAIN_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_CLIPCHILDREN;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_APPWINDOW;
			bMinimizeToTaskBar = (dwWinBinderStyle & WBC_COLLAPSED) != 0;
			break;

		case ModalDialog:			// Modal dialog box
			pszClass = MODAL_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_POPUP | WS_SYSMENU | WS_CAPTION | WS_VISIBLE | WS_CLIPCHILDREN | DS_MODALFRAME;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_DLGMODALFRAME;
			break;

		case ModelessDialog:		// Modeless dialog box
			pszClass = MODELESS_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPEDWINDOW | WS_SYSMENU | WS_CAPTION | WS_VISIBLE | WS_CLIPCHILDREN;
//			WS_POPUP | WS_SYSMENU | WS_CAPTION | WS_VISIBLE | WS_CLIPCHILDREN;
			dwExStyle = dwExStyle ? dwExStyle : 0;
			break;

		case ToolDialog:			// Modeless tool dialog
			pszClass = MODELESS_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_POPUP | WS_SYSMENU | WS_CAPTION | WS_VISIBLE | WS_CLIPCHILDREN;
			dwExStyle = dwExStyle ? dwExStyle : WS_EX_PALETTEWINDOW;
			break;

		case LayoutDialog:		// Layout dialog
			pszClass = LAYOUT_WINDOW_CLASS;
			dwStyle = dwStyle ? dwStyle : WS_OVERLAPPED | WS_SIZEBOX | WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_CLIPCHILDREN;
			dwExStyle = dwExStyle ? dwExStyle : 0;
			break;

		default:
			return NULL;
	}

	// Create the window

	if(BITTEST(dwWinBinderStyle, WBC_INVISIBLE))
		dwStyle = dwStyle & (~WS_VISIBLE);

	strncpy(szWindowName, pszCaption, MAXWINNAME - 1);
	pwbo->handle = CreateWindowEx(
		dwExStyle,
		pszClass,
		szWindowName,
		dwStyle ? dwStyle : DEFAULT_WIN_STYLE,
		(int)xPos < 0 ? CW_USEDEFAULT : xPos,
		(int)yPos < 0 ? CW_USEDEFAULT : yPos,
		nWidth <= 0 ? CW_USEDEFAULT : nWidth,
		nHeight <= 0 ? CW_USEDEFAULT : nHeight,
		pwboParent ? pwboParent->handle : NULL,
		NULL, hAppInstance, NULL);

	if(!pwbo->handle) {
		wbError(MB_ICONWARNING, "Error creating window");
		return NULL;
	}

	SetWindowLong(pwbo->handle, GWL_USERDATA, (LONG)pwbo);

	// Is it a modal dialog?

	if(uWinBinderClass == ModalDialog)
		SendMessage(pwbo->handle, WM_SETFOCUS, (WPARAM)GetWindow(pwbo->handle, GW_CHILD), 0);

	return pwbo;
}

BOOL wbSetWindowHandler(PWBOBJ pwbo, LPCTSTR pszHandler)
{
	HWND hChild = NULL;
	PWBOBJ poChild;

	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	if(!pszHandler || !*pszHandler)
		return FALSE;

	pwbo->pszCallBackFn = (LPTSTR)pszHandler;

	// Look for child tab controls to assign the same handler to them

	while((hChild = FindWindowEx(pwbo->handle, hChild, WC_TABCONTROL, NULL)) != NULL) {
		poChild = wbGetWBObj(hChild);
		poChild->pszCallBackFn = poChild->parent->pszCallBackFn;
	}

	return TRUE;
}

BOOL wbDestroyWindow(PWBOBJ pwbo)
{
	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	return DestroyWindow(pwbo->handle);
}

BOOL wbSetWindowSize(PWBOBJ pwbo, int nWidth, int nHeight)
{
	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	nWidth = MAX(0, nWidth);
	nHeight = MAX(0, nHeight);
	return SetWindowPos(pwbo->handle, 0, 0, 0, nWidth, nHeight, SWP_NOMOVE | SWP_NOZORDER);
}

/*
	TODO: Center horizontally only or vertically only according to xPos, yPos
*/

BOOL wbSetWindowPosition(PWBOBJ pwbo, int xPos, int yPos, PWBOBJ pwboParent)
{
	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	if(xPos == WBC_CENTER || yPos == WBC_CENTER) {
		if(!pwboParent || !pwboParent->handle || !IsWindow(pwboParent->handle))
			return wbCenterWindow(pwbo->handle, NULL);
		else
			return wbCenterWindow(pwbo->handle, pwboParent->handle);
	} else
		// pwboParent is ignored here
		return SetWindowPos(pwbo->handle, 0, xPos, yPos, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

/*
	TODO: Use pwboParent as the reference
*/

DWORD wbGetWindowPosition(PWBOBJ pwbo, PWBOBJ pwboParent)
{
	RECT rc;
	BOOL bRet;

	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	// pwboParent is ignored here
	bRet = GetWindowRect(pwbo->handle, &rc);
	if(!bRet)
		return (DWORD)MAKELONG(WBC_CENTER, WBC_CENTER);
	else
		return (DWORD)MAKELONG(rc.left, rc.top);
}

DWORD wbGetWindowSize(PWBOBJ pwbo)
{
	RECT rc;
	BOOL bRet;

	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	bRet = GetWindowRect(pwbo->handle, &rc);
	if(!bRet)
		return (DWORD)MAKELONG(WBC_CENTER, WBC_CENTER);
	else
		return (DWORD)MAKELONG(rc.right - rc.left, rc.bottom - rc.top);
}

/* Fills an array with all controls and returns the number of controls found. If pwboList is NULL,
  only returns the number of controls. */

UINT wbGetControlList(PWBOBJ pwboParent, PWBOBJ pwboList[], UINT nMaxControls)
{
	HWND hwnd;
	UINT i = 0;

	if(!wbIsWBObj((void *)pwboParent))
		return 0;

	if(!nMaxControls)
		return 0;

	hwnd = GetWindow(pwboParent->handle, GW_CHILD);
	if(hwnd) {
    	if(pwboList)
			pwboList[i] = wbGetWBObj(hwnd);
	} else
		return 0;
	for(i = 1; i < nMaxControls; i++) {
	    hwnd = GetWindow(hwnd, GW_HWNDNEXT);
	    if(hwnd) {
	    	if(pwboList)
	    		pwboList[i] = wbGetWBObj(hwnd);
	    } else
	    	break;
	}
	return i;
}

/* Sort a list view column */

BOOL wbSortLVColumn(PWBOBJ pwbo, int nSubItem, BOOL bAscending)
{
	hwndListView = pwbo->handle;		// For CompareLVItems()

	SendMessage(pwbo->handle, LVM_SORTITEMS, nSubItem, (bAscending ?
	    (LPARAM)(PFNLVCOMPARE)CompareLVItemsAscending :
	    (LPARAM)(PFNLVCOMPARE)CompareLVItemsDescending));
	UpdateLVlParams(hwndListView);
	return TRUE;
}

/* Make uPeriod equal to zero to kill the timer */

BOOL wbSetTimer(PWBOBJ pwbo, int id, UINT uPeriod)
{
	if(!pwbo || !pwbo->handle || !IsWindow(pwbo->handle))
		return FALSE;

	if(!uPeriod)
		return KillTimer(pwbo->handle, id);

	if(SetTimer(pwbo->handle, id, uPeriod, NULL))
		return TRUE;
	else
		return FALSE;
}

BOOL wbCenterWindow(HWND hwndDlg, HWND hwndParent)
{
	RECT rcDlg, rcOwner, rc;

	// From Windows Help, "Initializing a Dialog Box"

	// Get the owner window and dialog box rectangles.

	if(!hwndParent)
		hwndParent = GetDesktopWindow();
	GetWindowRect(hwndParent, &rcOwner);
	GetWindowRect(hwndDlg, &rcDlg);
	CopyRect(&rc, &rcOwner);

	// Offset the owner and dialog box rectangles so that
	// right and bottom values represent the width and
	// height, and then offset the owner again to discard
	// space taken up by the dialog box.

	OffsetRect(&rcDlg, -rcDlg.left, -rcDlg.top);
	OffsetRect(&rc, -rc.left, -rc.top);
	OffsetRect(&rc, -rcDlg.right, -rcDlg.bottom);

	// The new position is the sum of half the remaining
	// space and the owner's original position.

	SetWindowPos(hwndDlg,
		HWND_TOP,
		rcOwner.left + (rc.right / 2),
		rcOwner.top + (rc.bottom / 2),
		0, 0,          // ignore size arguments
		SWP_NOSIZE | SWP_NOZORDER);

	return TRUE;
}

//------------------------------------------- FUNCTIONS PUBLIC TO WINBINDER ONLY

void SetToolBarHandle(HWND hCtrl)
{
	hToolBar = hCtrl;
}

void SetStatusBarId(int id)
{
	idStatusBar = id;
}

/* Register a window class */

BOOL RegisterClasses(void)
{
	WNDCLASS wc;

	// Main window class

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = /*CS_HREDRAW | CS_VREDRAW | */CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)MainWndProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wc.lpszClassName = MAIN_WINDOW_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);	//IDI_WINLOGO
	wc.cbWndExtra = 0;

	if(!RegisterClass(&wc))
		return FALSE;

	// Modal dialogs

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)ModalWndProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wc.lpszClassName = MODAL_WINDOW_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = NULL;//LoadIcon(NULL, IDI_WINLOGO);
	wc.cbWndExtra = DLGWINDOWEXTRA;

	if(!RegisterClass(&wc))
		return FALSE;

	// Modeless dialogs

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)ModelessWndProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wc.lpszClassName = MODELESS_WINDOW_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = NULL;//LoadIcon(NULL, IDI_WINLOGO);
	wc.cbWndExtra = DLGWINDOWEXTRA;

	if(!RegisterClass(&wc))
		return FALSE;

	// Layout dialog

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)LayoutWndProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wc.lpszClassName = LAYOUT_WINDOW_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = NULL;
	wc.cbWndExtra = DLGWINDOWEXTRA;

	if(!RegisterClass(&wc))
		return FALSE;

	// Tab pages

	memset(&wc, 0, sizeof(WNDCLASS));
	wc.style = CS_DBLCLKS | CS_PARENTDC;
	wc.lpfnWndProc = (WNDPROC)TabPageProc;
	wc.hInstance = hAppInstance;
	wc.hbrBackground = hbrTabs ? hbrTabs : (HBRUSH)(COLOR_BTNFACE + 1);	// Not for under Windows XP
	wc.lpszClassName = TAB_PAGE_CLASS;
	wc.lpszMenuName = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hIcon = NULL;
	wc.cbWndExtra = DLGWINDOWEXTRA;

	if(!RegisterClass(&wc))
		return FALSE;

	return TRUE;
}

//------------------------------------------------------------ PRIVATE FUNCTIONS

/*
	The all-important common processing routine for all parent windows

	TODO: Filter user commands. Try: (LPNMHDR)lParam)->code == NM_CLICK, etc.
	TODO: Use the return value as a flag to know whether the calling function should continue default processing or not

*/

static LRESULT CALLBACK DefaultWBProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static HWND hTBWnd = NULL;			// Handle of toolbar window

	switch(msg) {

		case WM_NOTIFY:
			{
				char szClass[256];
				HWND hCtrl = ((LPNMHDR)lParam)->hwndFrom;
				PWBOBJ pwbobj;

				if(!IsWindow(hCtrl))
					break;

				// The classes below don't call the callback function

				GetClassName(hCtrl, szClass, 255);

				if(!stricmp(szClass, TOOLBARCLASSNAME)) {			// Toolbar

					if(!hTBWnd)
						hTBWnd = ((LPNMHDR)lParam)->hwndFrom;
					break;

				} else if(!stricmp(szClass, TOOLTIPS_CLASS)) {		// Tooltip

					if(((LPNMHDR)lParam)->code == (UINT)TTN_NEEDTEXT) {
						if(hTBWnd) {
							TBBUTTON tbb;
							int index;

							index = SendMessage(hTBWnd, TB_COMMANDTOINDEX,
							  ((LPTOOLTIPTEXT)lParam)->hdr.idFrom, 0);
							SendMessage(hTBWnd, TB_GETBUTTON, index, (LPARAM)(LPTBBUTTON)&tbb);
							((LPTOOLTIPTEXT)lParam)->lpszText = (LPTSTR)tbb.dwData;
						}
					}
					break;
				}

				// Verify if window is a WinBinder object

				pwbobj = wbGetWBObj(hCtrl);

				if(!pwbobj || !pwbobj->parent)
					break;

				if(!pwbobj->parent->pszCallBackFn)
					break;

				// Call callback function according to WinBinder class

				switch(pwbobj->uClass) {

					case Spinner:

						if(((LPNMHDR)lParam)->code == (UINT)UDN_DELTAPOS) {
							CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, 0);
						}
						break;

					case TreeView:

						switch(((LPNMHDR)lParam)->code) {

							case NM_DBLCLK:
								if(SEND_MESSAGE(WBC_DBLCLICK))
									CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, WBC_DBLCLICK);
								break;

							case (UINT)TVN_SELCHANGED:
								CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, 0);
								break;
						}
						break;

					case TabControl:

						if(((LPNMHDR)lParam)->code == (UINT)TCN_SELCHANGE) {

							HWND hTab = ((LPNMHDR)lParam)->hwndFrom;
							int nSelTab = TabCtrl_GetCurSel(hTab);

							wbSelectTab(wbGetWBObj(hCtrl), nSelTab);
							CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, nSelTab);
							return 0;
						}
						break;

					case ListView:

						switch(((LPNMHDR)lParam)->code) {

							case NM_DBLCLK:
								if(SEND_MESSAGE(WBC_DBLCLICK))
									CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, WBC_DBLCLICK);
								break;

							case LVN_ITEMCHANGED:

								if(((LPNM_LISTVIEW)lParam)->uChanged & (LVIF_STATE | LVIS_CHECKED)) {
									CALL_CALLBACK(((LPNMHDR)lParam)->idFrom, 0);
								}
								break;

							case LVN_COLUMNCLICK:

								hwndListView = pwbobj->handle;		// For CompareLVItems()
								SendMessage(pwbobj->handle, LVM_SORTITEMS,
									((NM_LISTVIEW FAR *)lParam)->iSubItem, (LPARAM)(PFNLVCOMPARE)CompareLVItemsAscending);
								UpdateLVlParams(hwndListView);
								break;
						}
						break;

				} // switch(pwbobj->uClass)
			} // ~WM_NOTIFY
			break;

		case WM_HSCROLL:		// Scroll bars, sliders
		case WM_VSCROLL:

			{
				PWBOBJ pwbobj;
				HWND hCtrl;
				int nMin, nMax, nPos;

				hCtrl = (HWND)lParam;
				pwbobj = wbGetWBObj(hCtrl);

				if(!pwbobj)
					break;

				nPos = SendMessage(hCtrl, SBM_GETPOS, 0, 0);

				switch(LOWORD(wParam)) {
					case SB_LINEUP:
						SendMessage(hCtrl, SBM_GETRANGE,
							(WPARAM)(LPINT)&nMin, (WPARAM)(LPINT)&nMax);
						nPos -= (nMax - nMin) / 100;
						nPos = MAX(nPos, 0);
						break;

					case SB_LINEDOWN:
						SendMessage(hCtrl, SBM_GETRANGE,
							(WPARAM)(LPINT)&nMin, (WPARAM)(LPINT)&nMax);
						nPos += (nMax - nMin) / 100;
						nPos = MAX(nPos, 0);
						break;

					case SB_PAGEUP:
						SendMessage(hCtrl, SBM_GETRANGE,
							(WPARAM)(LPINT)&nMin, (WPARAM)(LPINT)&nMax);
						nPos -= (nMax - nMin) / 10;
						nPos = MAX(nPos, 0);
						break;

					case SB_PAGEDOWN:
						SendMessage(hCtrl, SBM_GETRANGE,
							(WPARAM)(LPINT)&nMin, (WPARAM)(LPINT)&nMax);
						nPos += (nMax - nMin) / 10;
						nPos = MAX(nPos, 0);
						break;

					case SB_THUMBPOSITION:
					case SB_THUMBTRACK:						// Same value as TB_THUMBTRACK
						nPos = (short)HIWORD(wParam);
//						printf("%d ", LOWORD(wParam));
						break;
				}

				SendMessage(hCtrl, SBM_SETPOS, nPos, TRUE);
				CALL_CALLBACK(pwbobj->id, 0);
			}
			break;

		case WM_COMMAND:
			{
				HWND hCtrl;
				PWBOBJ pwbobj;

				if((HWND)lParam == 0 && (HIWORD(wParam) == 0 || HIWORD(wParam) == 1)) {		// Is the message from a menu or accelerator?

					pwbobj = wbGetWBObj(hwnd);
					// Must not use the macro CALL_CALLBACK here
					if(pwbobj && pwbobj->pszCallBackFn && *pwbobj->pszCallBackFn)
						wbCallUserFunction(pwbobj->pszCallBackFn, pwbobj, pwbobj, LOWORD(wParam), 0);
					return 0;
				}

				hCtrl = (HWND)lParam;
				pwbobj = wbGetWBObj(hCtrl);

				if(!pwbobj || !pwbobj->parent || !pwbobj->parent->pszCallBackFn)
					break;

				switch(pwbobj->uClass) {	// Classes that receive their notifications via WM_COMMAND

					case EditBox:
					case RTFEditBox:
						if((HIWORD(wParam) == EN_CHANGE)) {
							CALL_CALLBACK(LOWORD(wParam), 0);
						} else if(HIWORD(wParam) == EN_SETFOCUS) {
							if(SEND_MESSAGE(WBC_GETFOCUS))
								CALL_CALLBACK(LOWORD(wParam), WBC_GETFOCUS);
						}
						break;

					case ComboBox:
						if((HIWORD(wParam) == CBN_SELCHANGE)) {
							CALL_CALLBACK(LOWORD(wParam), 0);
						} else if(HIWORD(wParam) == CBN_SETFOCUS) {
							if(SEND_MESSAGE(WBC_GETFOCUS))
								CALL_CALLBACK(LOWORD(wParam), WBC_GETFOCUS);
						}
						break;

					case ListBox:
						if(HIWORD(wParam) == LBN_DBLCLK) {
							if(SEND_MESSAGE(WBC_DBLCLICK))
								CALL_CALLBACK(LOWORD(wParam), WBC_DBLCLICK);
						} else if (HIWORD(wParam) == LBN_SELCHANGE) {
							CALL_CALLBACK(LOWORD(wParam), 0);
						} else if(HIWORD(wParam) == LBN_SETFOCUS) {
							if(SEND_MESSAGE(WBC_GETFOCUS))
								CALL_CALLBACK(LOWORD(wParam), WBC_GETFOCUS);
						}
						break;

					case PushButton:
					case CheckBox:
					case RadioButton:
					case ToolBar:

						if(HIWORD(wParam) == BN_CLICKED) {
							CALL_CALLBACK(LOWORD(wParam), 0);
						} else if(HIWORD(wParam) == BN_SETFOCUS) {
							if(SEND_MESSAGE(WBC_GETFOCUS))
								CALL_CALLBACK(LOWORD(wParam), WBC_GETFOCUS);
						}
						break;

//					default:				// TODO: Maybe delete these lines,
//						bProcess = TRUE;	// but must be sure first
//						CALL_CALLBACK(LOWORD(wParam), 0);
				}

				return 0;
			}
			break;

		case WM_CTLCOLORSTATIC:			// For static controls and others
		case WM_CTLCOLORBTN:			// For pushbuttons

			if(hbrTabs) {				// Not for versions under Windows XP
				HWND hCtrl;
				PWBOBJ pwbobj;

				hCtrl = (HWND)lParam;
				pwbobj = wbGetWBObj(hCtrl);

				if(!pwbobj || !pwbobj->parent)
					break;

				if(pwbobj->uClass == EditBox)				// Not for edit controls
					break;

				SetBkColor((HDC)wParam, clrTabs);			// Static controls need this
				if(pwbobj->parent->uClass == TabControl)	// Only for controls on tabs
					return (LRESULT)hbrTabs;				// Paint the background with the tab page color
			}
			break;

		case WM_TIMER:

			{
				PWBOBJ pwbobj;

				pwbobj = wbGetWBObj(hwnd);

				if(!pwbobj || !pwbobj->pszCallBackFn)
					break;

				if(pwbobj->pszCallBackFn)
					wbCallUserFunction(pwbobj->pszCallBackFn, pwbobj, pwbobj, wParam, 0);
				return 0;
			}
			break;

	}

	return DefWindowProc(hwnd, msg, wParam, lParam);
}

static LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {

		case WM_CREATE:
			if(((LPCREATESTRUCT)lParam)->x == WBC_CENTER && ((LPCREATESTRUCT)lParam)->x == WBC_CENTER)
				wbCenterWindow(hwnd, NULL);
			break;

		case WM_ACTIVATE:
			hCurrentDlg = hwnd;	// Used in IsDialogMessage() -- main loop
			break;

		case WM_SIZE:
			if(bMinimizeToTaskBar && (wParam == SIZE_MINIMIZED)) {
				AddTaskBarIcon(hwnd);
				ShowWindow(hwnd, SW_HIDE);
			}
			SendDlgItemMessage(hwnd, idStatusBar, msg, wParam, lParam);
			SendMessage(hToolBar, msg, wParam, lParam);
			break;

		case WM_CLOSE:
			{
				PWBOBJ pwbobj = (PWBOBJ)GetWindowLong(hwnd, GWL_USERDATA);

				if(pwbobj && pwbobj->pszCallBackFn && *pwbobj->pszCallBackFn) {

					// This message is only sent to windows that have callback functions

					SendMessage(hwnd, WM_COMMAND, MAKELONG(IDCLOSE, 0), (LPARAM)NULL);
					return 0;
				} else {
		 			break;
				}
 			}
			break;

		case WBWM_NOTIFYICON:
			if(bMinimizeToTaskBar) {

				switch (lParam){

					case WM_LBUTTONDBLCLK:
					case WM_LBUTTONUP:

						ShowWindow(hwnd, SW_SHOWMINNOACTIVE);
						ShowWindow(hwnd, SW_NORMAL);
						SetForegroundWindow(hwnd);
						SetFocus(hwnd);
						DeleteTaskBarIcon(hwnd);
						return TRUE;
				}
			}
			break;

		case WM_DESTROY:
			if(bMinimizeToTaskBar)
				DeleteTaskBarIcon(hwnd);
			PostQuitMessage(0);
			break;

		default:
			return DefaultWBProc(hwnd, msg, wParam, lParam);
	}
	return DefWindowProc(hwnd, msg, wParam, lParam);
}

// The word "modal" here is not quite true: these are not "real" modal dialog boxes

static LRESULT CALLBACK ModalWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {

		case WM_CREATE:
			if(((LPCREATESTRUCT)lParam)->x == WBC_CENTER && ((LPCREATESTRUCT)lParam)->x == WBC_CENTER)
				wbCenterWindow(hwnd, NULL);
			EnableWindow(GetParent(hwnd), FALSE);
			break;

		case WM_ACTIVATE:
			hCurrentDlg = hwnd;	// Used in IsDialogMessage() -- main loop
			break;

		case WM_CLOSE:
			{
				HWND hParent;

				hParent = GetParent(hwnd);
				if(hParent) {
					EnableWindow(hParent, TRUE);
					SetFocus(hParent);
				}
				DestroyWindow(hwnd);
				hwnd = NULL;
			}
			break;

		case WM_DESTROY:
			EnableWindow(GetParent(hwnd), TRUE);
			SetFocus(GetParent(hwnd));
			hCurrentDlg = NULL;
			break;

		default:
			return DefaultWBProc(hwnd, msg, wParam, lParam);
	}
	return DefDlgProc(hwnd, msg, wParam, lParam);
}

static LRESULT CALLBACK ModelessWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {

		case WM_CREATE:
			if(((LPCREATESTRUCT)lParam)->x == WBC_CENTER && ((LPCREATESTRUCT)lParam)->x == WBC_CENTER)
				wbCenterWindow(hwnd, NULL);
			EnableWindow(GetParent(hwnd), FALSE);
			break;

		case WM_ACTIVATE:
			hCurrentDlg = hwnd;	// Used in IsDialogMessage() -- main loop
			break;

		case WM_CLOSE:
			{
				HWND hParent;

				hParent = GetParent(hwnd);
				if(hParent) {
					EnableWindow(hParent, TRUE);
					SetFocus(hParent);
				}
				DestroyWindow(hwnd);
				hwnd = NULL;
			}
			break;

		case WM_DESTROY:
			EnableWindow(GetParent(hwnd), TRUE);
			SetFocus(GetParent(hwnd));
			hCurrentDlg = NULL;
			break;

		default:
			return DefaultWBProc(hwnd, msg, wParam, lParam);
	}
	return DefDlgProc(hwnd, msg, wParam, lParam);
}

// Processes messages for tab pages

static LRESULT CALLBACK TabPageProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg) {

		case WM_CLOSE:		// Prevents that an ESC from inside an edit box closes this page
			return 0;

		case WM_SHOWWINDOW:
			if(wParam) {
				HWND hwndNext;

				hCurrentDlg = hwnd;	// Used in IsDialogMessage() -- main loop
				hwndNext = GetNextDlgTabItem(hwnd, NULL, FALSE);
				if(hwndNext)
					SetFocus(hwndNext);
			} else
				hCurrentDlg = NULL;
			break;
	}
	return DefaultWBProc(hwnd, msg, wParam, lParam);
}

// Adapted from from http://groups.google.com/groups?hl=en&lr=&selm=38BC4F60.11F62F%40thematic.com
// Peter Forbes / David Lowndes

// Try several methods to retrieve the icon from an application window

static HICON GetWindowIcon(HWND hwnd)
{
	HICON hIcon;

	hIcon = (HICON)SendMessage(hwnd, WM_GETICON, ICON_SMALL, 0);
	if(hIcon)
		return hIcon;

	hIcon = (HICON)SendMessage(hwnd, WM_GETICON, ICON_BIG, 0);
	if(hIcon)
		return hIcon;

	hIcon = (HICON)SendMessage(hwnd, WM_QUERYDRAGICON, 0, 0);
	if(hIcon)
		return hIcon;

	hIcon = (HICON)GetClassLong(hwnd, GCL_HICONSM);
	if(hIcon)
		return hIcon;

	hIcon = (HICON)GetClassLong(hwnd, GCL_HICON);
	if(hIcon)
		return hIcon;

	return LoadIcon(NULL, IDI_APPLICATION);
}

// From Win32 programmer's reference, "Adding and Deleting Taskbar Icons" and
// http://msdn.microsoft.com/library/en-us/shellcc/platform/shell/programmersguide/shell_int/shell_int_programming/taskbar.asp

static BOOL AddTaskBarIcon(HWND hwnd)
{
    BOOL bRes;
	NOTIFYICONDATA tnid;

	if(!hwnd)
		return FALSE;

	tnid.cbSize = sizeof(NOTIFYICONDATA);
	tnid.hWnd = hwnd;
	tnid.uID = 0;
	tnid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
	tnid.uCallbackMessage = WBWM_NOTIFYICON;
	tnid.hIcon = CopyIcon(GetWindowIcon(hwnd));
	GetWindowText(hwnd, tnid.szTip, 63);		// szTip[] is 64 bytes long
	bRes = Shell_NotifyIcon(NIM_ADD, &tnid);
	if(tnid.hIcon)
		DestroyIcon(tnid.hIcon);
	return bRes;
}

// From Win32 programmer's reference, "Adding and Deleting Taskbar Icons" and
// http://msdn.microsoft.com/library/en-us/shellcc/platform/shell/programmersguide/shell_int/shell_int_programming/taskbar.asp

static BOOL DeleteTaskBarIcon(HWND hwnd)
{
	BOOL bRes;
	NOTIFYICONDATA tnid;

	if(!hwnd)
		return FALSE;
	tnid.cbSize = sizeof(NOTIFYICONDATA);
	tnid.hWnd = hwnd;
	tnid.uID = 0;
	bRes = Shell_NotifyIcon(NIM_DELETE, &tnid);
	DestroyIcon(tnid.hIcon);
	return bRes;
}

/*

  Ported from Iczelion's Win32 Assembly Homepage: http://spiff.tripnet.se/~iczelion/tut31.html

  Quote: "Note that when the items are rearranged, their indexes also change. So when the sorting is done
  in my example, I need to update the values in lParam to reflect the new indexes."

*/

static void UpdateLVlParams(HWND hwnd)
{
	LV_ITEM lvi;
	int count;

	count = SendMessage(hwnd, LVM_GETITEMCOUNT, 0, 0);
	lvi.mask = LVIF_PARAM;
	lvi.iSubItem = 0;
	for(lvi.iItem = 0; lvi.iItem < count; lvi.iItem++) {
		lvi.lParam = lvi.iItem;
		SendMessage(hwnd, LVM_SETITEM, 0, (LPARAM)&lvi);
	}
}

/*

 This is the callback function that compares two ListView items.

 - lParam1 && lParam2 contain the item indices
 - lParamSort is the sub-item index
 - 64 characters is more than enough for string comparison

*/

static BOOL IsDecimalNumber(LPTSTR szStr)
{
	int i;
	int nLen = MIN(strlen(szStr), 63);

	for(i = 0; i < nLen; i++)
		if((*(szStr + i) < '0' || *(szStr + i) > '9') && *(szStr + i) != ' ')
			return FALSE;
	return TRUE;
}

static int CALLBACK CompareLVItemsAscending(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	char szStr1[64];
	char szStr2[64];

	ListView_GetItemText(hwndListView, lParam1, lParamSort, szStr1, 63);
	ListView_GetItemText(hwndListView, lParam2, lParamSort, szStr2, 63);

	if(!*szStr1 && !*szStr2)				// Two empty strings
		return 0;
	else if(!*szStr2)						// Second string is empty
		return 1;
	else if(!*szStr1)						// First string is empty
		return -1;
	else if(IsDecimalNumber(szStr1) && IsDecimalNumber(szStr2))		// Two integers
		return(atol(szStr1) == atol(szStr2) ? 0 : ((atol(szStr1) > atol(szStr2)) ? 1 : -1));
	else
		return lstrcmp(szStr1, szStr2);		// According to locale, not case-sensitive
}

static int CALLBACK CompareLVItemsDescending(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	char szStr1[64];
	char szStr2[64];

	ListView_GetItemText(hwndListView, lParam1, lParamSort, szStr1, 63);
	ListView_GetItemText(hwndListView, lParam2, lParamSort, szStr2, 63);

	if(!*szStr1 && !*szStr2)				// Two empty strings
		return 0;
	else if(!*szStr2)						// Second string is empty
		return -1;
	else if(!*szStr1)						// First string is empty
		return 1;
	else if(IsDecimalNumber(szStr1) && IsDecimalNumber(szStr2))		// Two integers
		return(atol(szStr1) == atol(szStr2) ? 0 : ((atol(szStr1) > atol(szStr2)) ? -1 : 1));
	else
		return lstrcmp(szStr2, szStr1);		// According to locale, not case-sensitive
}

//------------------------------------------------- CODE FOR LAYOUT WINDOW CLASS

#define NO_VALUE		(-65535)

int snapX = 8;
int snapY = 8;

/// Thanks to Dave Edson

typedef struct DIALOGHANDLEtag {
	RECT   rcArea;             // Handle's coordinates
	DWORD  dwNCHITVALUE;       // NCHITTEST return value
} DIALOGHANDLE;

DIALOGHANDLE DialogHandles[8]; // Handles for each area to grab

void ScreenToClientRect(HWND hwnd, LPRECT prRect)
{
	POINT pt;

	pt.x = prRect->left;
	pt.y = prRect->top;
	ScreenToClient(hwnd, &pt);
	prRect->left = pt.x;
	prRect->top  = pt.y;
	pt.x = prRect->right;
	pt.y = prRect->bottom;
	ScreenToClient(hwnd, &pt);
	prRect->right = pt.x;
	prRect->bottom = pt.y;
}

#define MAKEHANDLE(iIndex,x,y,dwHitValue)           \
{                                                   \
  DialogHandles[iIndex].rcArea.left   = x;          \
  DialogHandles[iIndex].rcArea.top    = y;          \
  DialogHandles[iIndex].rcArea.right  = x + dx + 1; \
  DialogHandles[iIndex].rcArea.bottom = y + dy + 1; \
  DialogHandles[iIndex].dwNCHITVALUE  = dwHitValue; \
}

void FigureHandleSizes(HWND hwnd)
{
	RECT rc;
	int  dx = GetSystemMetrics(SM_CXFRAME);
	int  dy = GetSystemMetrics(SM_CYFRAME) + 1;

	GetWindowRect(hwnd, &rc);

	MAKEHANDLE( 0, rc.left,                   (rc.top+rc.bottom)/2-dy/2, HTLEFT          )
	MAKEHANDLE( 1, rc.right-dx,               (rc.top+rc.bottom)/2-dy/2, HTRIGHT         )
	MAKEHANDLE( 2, (rc.left+rc.right)/2-dx/2, rc.top,                    HTTOP           )
	MAKEHANDLE( 3, rc.left,                   rc.top,                    HTTOPLEFT       )
	MAKEHANDLE( 4, rc.right-dx,               rc.top,                    HTTOPRIGHT      )
	MAKEHANDLE( 5, (rc.left+rc.right)/2-dx/2, rc.bottom-dy,              HTBOTTOM        )
	MAKEHANDLE( 6, rc.left,                   rc.bottom-dy,              HTBOTTOMLEFT    )
	MAKEHANDLE( 7, rc.right-dy,               rc.bottom-dy,              HTBOTTOMRIGHT   )
}

// Processing routine for resource editor controls

static LRESULT CALLBACK CtrlProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	POINT pt;
	RECT rc;
	int i;
	static POINT ptOffset = {NO_VALUE, NO_VALUE};
	static POINT ptDown = {NO_VALUE, NO_VALUE};
	static BOOL bIsMoving = FALSE;

	switch(msg) {

		case WM_NCHITTEST:		// Passes commands to parent window

			FigureHandleSizes (hwnd);

			GetClientRect (hwnd, &rc);
			pt.x = (int) LOWORD (lParam);
			pt.y = (int) HIWORD (lParam);

			// First, test for the "handles"

			for(i = 0; i < 8; i++)
				if(PtInRect(&(DialogHandles[i].rcArea), pt))
					return DialogHandles[i].dwNCHITVALUE;

			// Second, test for moving

			ScreenToClient(hwnd, &pt);
			if(PtInRect(&rc, pt)) {
				bIsMoving = TRUE;
				return HTCAPTION;
			} else
				break;

		case WM_NCLBUTTONDBLCLK:		// Prevents the control from maximizing itself
			return 0;

		case WM_WINDOWPOSCHANGING:		// Prevents the control from improperly surfacing to the top of the Z-order
			if(bIsMoving) {
				((LPWINDOWPOS)lParam)->flags = ((LPWINDOWPOS)lParam)->flags | (SWP_NOZORDER);
				bIsMoving = FALSE;
			}
			return 0;

		case WM_NCLBUTTONDOWN:
			SetFocus(hwnd);
			GetCursorPos(&ptDown);
			break;

		case WM_WINDOWPOSCHANGED:		// Sends messages to parent handler
			{
				PWBOBJ pwbobj;

				pwbobj = wbGetWBObj(hwnd);

				if(pwbobj && pwbobj->parent && pwbobj->parent->pszCallBackFn && *pwbobj->parent->pszCallBackFn) {
					wbCallUserFunction(pwbobj->parent->pszCallBackFn, pwbobj->parent, pwbobj, pwbobj->id, 0);
				}
				return 0;
			}
			break;

		case WM_MOVING:
			{
				POINT pt;
				RECT *pRect = (RECT *)lParam;
				static SIZE szSize;

				// Gets done only once

				if(ptDown.x != NO_VALUE && ptDown.y != NO_VALUE) {
					szSize.cx = pRect->right - pRect->left;
					szSize.cy = pRect->bottom - pRect->top;
					ptOffset.x = ptDown.x - pRect->left;
					ptOffset.y = ptDown.y - pRect->top;
					ptDown.x = NO_VALUE;
					ptDown.y = NO_VALUE;
				}

				GetCursorPos(&pt);

				pRect->left = (pt.x - ptOffset.x) / snapX * snapX;
				pRect->top = (pt.y - ptOffset.y) / snapY * snapY;
				pRect->right = pRect->left + szSize.cx;
				pRect->bottom = pRect->top + szSize.cy;
			}
			return TRUE;

		case WM_SIZING:
			{
				POINT pt;
				RECT *pRect = (RECT *)lParam;

				// Gets done only once

				if(ptDown.x != NO_VALUE && ptDown.y != NO_VALUE) {
					ptOffset.x = ptDown.x - pRect->left;
					ptOffset.y = ptDown.y - pRect->top;
					ptDown.x = NO_VALUE;
					ptDown.y = NO_VALUE;
				}

				GetCursorPos(&pt);

				switch(wParam) {
					case WMSZ_BOTTOM:
						pRect->bottom = pRect->top + (pt.y - pRect->top) / snapY * snapY;
						break;
					case WMSZ_BOTTOMLEFT:
						pRect->bottom = pRect->top + (pt.y - pRect->top) / snapY * snapY;
						pRect->left = (pt.x - ptOffset.x) / snapX * snapX;
						break;
					case WMSZ_BOTTOMRIGHT:
						pRect->bottom = pRect->top + (pt.y - pRect->top) / snapY * snapY;
						break;
					case WMSZ_LEFT:
						pRect->left = (pt.x - ptOffset.x) / snapX * snapX;
						break;
					case WMSZ_RIGHT:
						pRect->right = pRect->left + (pt.x - pRect->left) / snapX * snapX;
						break;
					case WMSZ_TOP:
						pRect->top = (pt.y - ptOffset.y) / snapY * snapY;
						break;
					case WMSZ_TOPLEFT:
						pRect->top = (pt.y - ptOffset.y) / snapY * snapY;
						pRect->left = (pt.x - ptOffset.x) / snapX * snapX;
						break;
					case WMSZ_TOPRIGHT:
						pRect->top = (pt.y - ptOffset.y) / snapY * snapY;
						pRect->right = pRect->left + (pt.x - pRect->left) / snapX * snapX;
						break;
				}
			}
			return TRUE;

		case WM_PAINT:

			// Let the control draw itself first

			CallWindowProc((WNDPROC)GetClassLong(hwnd, GCL_WNDPROC), hwnd, msg, wParam, lParam);

			// Draw Handles

			if(GetFocus() == hwnd) {
				HDC hDC;

				FigureHandleSizes(hwnd);
				hDC = GetWindowDC(hwnd);
				for(i = 0; i < 8; i++) {
					rc = DialogHandles[i].rcArea;
					ScreenToClientRect(hwnd, &rc);
					FillRect(hDC, &rc, GetStockObject(WHITE_BRUSH));
					FrameRect(hDC, &rc, GetStockObject(BLACK_BRUSH));
				}
				ReleaseDC(hwnd, hDC);
			}
			return 0;
	}

	// Calls original class procedure

	return CallWindowProc((WNDPROC)GetClassLong(hwnd, GCL_WNDPROC), hwnd, msg, wParam, lParam);
}

static LRESULT CALLBACK LayoutWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
//	printf("%d\n", msg);
	switch(msg) {

		case WM_CREATE:
			if(((LPCREATESTRUCT)lParam)->x == WBC_CENTER && ((LPCREATESTRUCT)lParam)->x == WBC_CENTER)
				wbCenterWindow(hwnd, NULL);
			EnableWindow(GetParent(hwnd), FALSE);
			break;

		case WM_PARENTNOTIFY:

			// When creating a child control, subclass it

			if(LOWORD(wParam) == WM_CREATE) {
				DWORD dwStyle;

				SetWindowLong((HWND)lParam, GWL_WNDPROC, (LONG)CtrlProc);
				dwStyle = GetWindowLong((HWND)lParam, GWL_STYLE);
				SetWindowLong((HWND)lParam, GWL_STYLE, dwStyle | WS_CLIPSIBLINGS);
			}
			break;

		case WM_PAINT:

			// Draw the grid dots
			{
				HDC hdc;
				RECT rc;
				PAINTSTRUCT ps;
				int x, y;

				GetClientRect(hwnd, &rc);
				hdc = BeginPaint(hwnd, &ps);
				for(y = 0; y < rc.bottom; y += snapX)
					for(x = 0; x < rc.right; x += snapY)
						SetPixel(hdc, x, y, RGB(0,0,0));
				EndPaint(hwnd, &ps);
			}
			return 0;

		case WM_CLOSE:						// Hide the window, don't destroy it
			ShowWindow(hwnd, SW_HIDE);
			break;

		case WM_DESTROY:
			EnableWindow(GetParent(hwnd), TRUE);
			SetFocus(GetParent(hwnd));
			hCurrentDlg = NULL;
			break;

//		default:
//			return DefaultWBProc(hwnd, msg, wParam, lParam);
	}
	return DefDlgProc(hwnd, msg, wParam, lParam);
}

//------------------------------------------------------------------ END OF FILE
